import os
import traceback
from colorama import Fore, Style, init
import datetime
import shutil
init(convert=True)

class BaseTest:
    """Creates the structure for test suites written using the Moonrise framework
    """

    # The dictionary to be referenced when evaluating what tests to run
    tests = {}


    def __init__(self, test_cases=()):
        """Evaluates whether to run tests within the current test suite based on user input and creates report folder and file structures.

           Arguments:
           - `test_cases`: Test Case names to be executed. If none specified, will execute all test cases under the current test suite.
        """

        self.suite_tests = self.tests.get(f"{self.__module__}.{self.__class__.__name__}")
        self.video_folder = None

        # If the condition is present where there are no tests in a suite or the user-requested tests cases do not match the current suite,
        # do not proceed.
        if self.suite_tests is None or (test_cases != () and len(set(test_cases).intersection(self.suite_tests)) == 0):
            # Add an extra note if test cases were specified but do not exist in the current suite.
            if test_cases and self.suite_tests:
                print(f'Skipping test suite "{self.__class__.__name__}" in module "{self.__module__}" because no tests matching {test_cases} were found.')
            return
        
        if test_cases:
            test_cases = set(test_cases).intersection(self.suite_tests)
        else:
            test_cases = self.suite_tests

        self.passes = 0
        self.failures = 0
        self.totals = 0

        self.colors = {
        "pass": Fore.LIGHTGREEN_EX,
        "fail": Fore.LIGHTRED_EX,
        "header": Fore.LIGHTCYAN_EX,
        "info": ""
        }

        if not os.path.exists(str(f"{os.getcwd()}/reports/{self.__module__}/{self.__class__.__name__}")):
            os.makedirs(str(f"{os.getcwd()}/reports/{self.__module__}/{self.__class__.__name__}"))
        self.reports_folder = str(f"{os.getcwd()}/reports/{self.__module__}/{self.__class__.__name__}")

        if not os.path.exists(str(f"{self.reports_folder}/video")):
            os.makedirs(str(f"{self.reports_folder}/video"))
        self.video_folder = self.reports_folder + "/video"

        self.report_file = open(f"{self.reports_folder}/{self.__class__.__name__}.log", "w")

        self.run_tests(test_cases)

        shutil.rmtree(self.video_folder)

    def run_tests(self, test_cases):
        """Method to call the requested tests.

           Arguments:
           - `test_cases`: Test Case names to be executed.
        """
        if self.video_thread:
            self.video_thread.video_folder = self.video_folder
        
        self.log_to_report(f"----------------- Beginning Suite: {self.__class__.__name__} -----------------", log_type="header")
        # Perform suite setup actions before any tests are executed.
        self.suite_setup()
        
        self.totals += len(test_cases)
        
        for tc_number, tc in enumerate(test_cases):
            self.log_to_report(f"--- Starting test {tc_number + 1} of {len(test_cases)}: {tc} ---", log_type="header")
            self.suite_tests.get(tc)(self)

        self.log_to_report(f"----------------- Ending Suite: {self.__class__.__name__} -----------------", log_type="header")
        # Perform suite teardown actions after all tests are executed.
        self.suite_teardown()

        if self.failures > 0:
            end_string = f"{self.colors.get('pass')}{self.passes} tests passing, {self.colors.get('fail')}{self.failures} tests failing, {self.colors.get('header')}{self.totals} tests total"
        else:
            end_string = f"{self.colors.get('pass')}{self.passes} tests passing, {self.colors.get('header')}{self.totals} tests total"

        self.log_to_report(end_string, log_type="header")

        # Create video from screenshots if video_thread was created.
        if self.video_thread:
            self.video_thread.create_video_from_pngs(f"{self.reports_folder}/{self.__class__.__name__}.mp4")
            self.video_thread.stop()

    def log_to_report(cls, message, log_type = "info"):
        """Log information with a timestamp to the console and to the report file.

           Arguments:
           - message: The text to log.
           - log_type: The color of the text as it will appear in the console. Consult the colors variable in this class for specifics.
        """
        timestamp = datetime.datetime.now()
        print(f"\n{cls.colors.get(log_type)}{timestamp} | {message}{Style.RESET_ALL}")
        for color in cls.colors.values():
            message = message.replace(color, "")
        cls.report_file.write(f"\n\n{timestamp} | {message}")

    def suite_setup(self):
        """Default suite setup method to be called before any tests are executed.
        """
        pass

    def suite_teardown(self):
        """Default suite teardown method to be called after all tests are executed.
        """
        pass

    def test_teardown(self):
        """Default test teardown method to be called after every test is executed.
        """
        pass

    def test_setup(self):
        """Default test setup method to be called before every tests is executed.
        """
        pass

    @classmethod
    def test(cls, test_case):
        """Decorator method to hang on every method that is to be classified as a test case.
           Adds test case methods to the tests variable in this class. 
           These tests will be referenced upon attempting to run their test suites.

           Arguments:
           - test_case: The name of the method to add as a test case.
        """

        def test_executor(self):
            """Executes the current test case, performing test setup actions, handling failures, and finally calling test teardown actions.
            """
            try:
                self.test_setup()
                test_case(self)
                self.log_to_report(f"{test_case.__name__} PASS", log_type = "pass")
                self.passes += 1
            except Exception:
                self.log_to_report(f"{traceback.format_exc()}")
                if self.moon_driver:
                    self.moon_driver.save_screenshot(f"{self.reports_folder}/{test_case.__name__}.png")
                self.log_to_report(f"{test_case.__name__} FAIL", log_type = "fail")
                self.failures += 1
            finally:
                self.test_teardown()

        mod_and_suite = f"{test_case.__module__}.{test_case.__qualname__.split('.')[0]}"
        cls.tests.setdefault(mod_and_suite, {})
        cls.tests[mod_and_suite][test_case.__name__] = test_executor
        return test_executor


import os
import subprocess
from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver
from selenium.webdriver.common.service import Service
from moonrise.Moon_Movie import VideoThread



class MoonBrowser:

    moon_driver = None
    video_thread = None

    # Default time to wait for elements to become visible.
    # Can be changed in a Moonrise Test Suite by setting Moonrise.default_timeout = (new timeout)
    default_timeout = 30

    def open_browser(self, browser_type, *browser_args, persist=False, record_test=True, shutter_speed=0.05):
        """Opens a selenium browser of a specified browser type
           Arguments:
           - browser_type: The desired browser (Chrome, Firefox, Edge, or IE).
           - browser_args: Selenium browser arguments, e.g. --headless.
           - persist: If set to True, will keep the browser open for later use.
           - record_test: If set to True, will create a video recording during the time that the browser is open.
           - shutter_speed: The delay in seconds between screenshots taken for the purpose of creating the video recording.
           Default is 0.05 seconds, or one screenshot every 20th of a second.
           Lower values wil result in more detailed videos, but larger performance hits.

           Creates class variable moon_driver for access to selenium webdriver methods.
        """ 

        browser_options = {
            'edge': {
                'options': webdriver.EdgeOptions(),
                'webdriver_create': webdriver.Edge
            },
            'chrome': {
                'options': webdriver.ChromeOptions(),
                'webdriver_create': webdriver.Chrome
            },
            'firefox': {
                'options': webdriver.FirefoxOptions(),
                'webdriver_create': webdriver.Firefox
            },
            'ie': {
                'options': webdriver.IeOptions(),
                'webdriver_create': webdriver.Ie
            },
        }

        if browser_type.lower() not in browser_options:
            raise KeyError(f"'{browser_type}' not in list of acceptable browsers. Acceptable browsers are chrome, edge, firefox, and ie")

        options = browser_options[browser_type]['options']

        for arg in browser_args:
            options.add_argument(arg)

        # Prevent the default browser cleanup 
        if persist == True:
            Service.__del__ = lambda new_del: None

        # moon_driver not only creates a browser session, but also can be used in higher-order methods to access selenium methods, e.g. refresh(), maximize_window(), etc.
        self.moon_driver = browser_options[browser_type]['webdriver_create'](options=options)

        # The WebDriverWait object that will inform how long elements are waited for.
        self.wait = WebDriverWait(self.moon_driver, self.default_timeout)

        # Creates VideoThread object.
        if persist != True and record_test == True:
            # Stops any previously running screenshot threads
            if self.video_thread:
                self.video_thread.stop()
            self.video_thread = VideoThread(self.moon_driver, shutter_speed)
            if self.video_folder:
                self.video_thread.video_folder = self.video_folder

        # write executor_url and session_id to a file named session_info.py for future use
        try:
            session_info_file = open(os.path.dirname(os.path.realpath(__file__))+'/session_info.py', 'w')
            session_info_file.write(f'executor_url="{self.moon_driver.command_executor._url}"\nsession_id="{self.moon_driver.session_id}"')
            session_info_file.close()
        except FileNotFoundError:
            pass

    def use_current_browser(self):
        """Allows for test executions to begin on the last opened browser
           Also creates the moon_driver variable for access to selenium webdriver methods
        """

        try:
            from moonrise import session_info
        except ImportError:
            raise ImportError("No open browser sessions.")

        # To prevent a new browser session from being created, we need to temporarily overwrite the RemoteWebDriver.execute method
        # Save the original function, so we can revert our patch
        org_command_execute = RemoteWebDriver.execute

        def new_command_execute(self, command, params=None):
            if command == "newSession":
                # Mock the response
                return {"value": {"sessionId": session_info.session_id, "capabilities": params}}
            else:
                return org_command_execute(self, command, params)

        # Patch the function before creating the driver object
        RemoteWebDriver.execute = new_command_execute

        self.moon_driver = webdriver.Remote(command_executor=session_info.executor_url, options=[])

        self.wait = WebDriverWait(self.moon_driver, self.default_timeout)

        # Replace the patched function with original function
        RemoteWebDriver.execute = org_command_execute


    def cleanup_browser(self):
        """Attempts to tear down most recent browser.
           Kills all geckodriver.exe, chromedriver.exe, and msedgedriver.exe processes.
        """
        if self.video_thread:
            self.video_thread.stop()

        try:
            self.moon_driver.quit()
            self.moon_driver = None
        except AttributeError:
            pass

        try:
            os.remove(os.path.dirname(os.path.realpath(__file__))+'/session_info.py')
        except FileNotFoundError:
            pass
        
        processes = ["geckodriver.exe", "chromedriver.exe", "msedgedriver.exe"]
        try:
            for process in processes:
                subprocess.call(f'taskkill /f /im {process}', stdout=open(os.devnull, "wb"), stderr=open(os.devnull, "wb"))
        except FileNotFoundError:
            pass
            
    def navigate_to_page(self, url):
        """Attempts to navigate to a web page without first needing https or http prefix
           Arguments:
           - url: The desired url
        """
        
        if not url.startswith("https") and not url.startswith("http"):
            url = "https://" + url

        self.moon_driver.get(url)



import os
import sys
from moonrise.Base_Test import BaseTest
from moonrise import Moonrise
from colorama import Fore, Style

def run_cli(args=None):
    """Method for running command line interface

       Arguments:
       - args: arguments from the command line specifying the python module or folder, the tests to run, and/or the suites to run.
    """
    if args is None:
        args = sys.argv[1:]
    
    def import_modules(filepath):
        """Add relevant modules to path and to modules to examine for Moonrise tests and suites.
        """

        if ":" in filepath:
            dir_name = filepath
        else:
            dir_name = os.getcwd() + "/" + filepath

        if os.path.isfile(dir_name):
            sys.path.append(os.path.abspath(os.path.join(dir_name, os.pardir)))
            module = os.path.basename(dir_name[:-3])
            modules.append(__import__(module))
        elif os.path.isdir(dir_name):
            sys.path.append(dir_name)
            for name in os.listdir(dir_name):
                if name.endswith(".py"):
                    module = name[:-3]    
                    modules.append(__import__(module))

    modules = []
    suites = []
    tests = []

    for arg in args:
        if arg.startswith("suite"):
            suites.append(arg.split(":")[1])
        elif arg.startswith("test"):
            tests.append(arg.split(":")[1])
        else:
            import_modules(arg)

    tests = tuple(tests)

    if not modules:
        print(f"{Fore.LIGHTRED_EX}moonrise Error: No valid file or directory given{Style.RESET_ALL}")
        return
    
    tests_were_run = False

    for module in modules:
        for clazz in module.__dict__.values():
            if BaseTest.__name__ in str(clazz.__init__) and clazz.__name__ != Moonrise.__name__:
                if suites:
                    if clazz.__name__ in suites:
                        tests_were_run = True
                        clazz(tests)
                else:
                    tests_were_run = True
                    clazz(tests)

    if not tests_were_run:
        print(f"{Fore.LIGHTRED_EX}moonrise Error: No tests were run. Were the tests or suites labeled correctly?{Style.RESET_ALL}")
        return


if __name__ == "__main__":
    run_cli(sys.argv[1:])


from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.support.ui import Select
from selenium.common.exceptions import TimeoutException

class MoonMethods:
    """Common-use methods when locating and interacting with web elements
    """

    def get_web_element(self, locator, timeout=None, get_multiples = False):
        """Locate a web element or elements via a given ``locator``.
           Throws a TimeoutException if the element(s) cannot be found within ``timeout`` or default_timeout

           Can perform a search in 3 main ways:
           - XPATH default: using the prefix '//' will automatically use XPATH to perform the search
           - CSS default: simply inputting an element locator will try to locate the element via CSS selector
           - By methods: specifying the lookup type will use the desired lookup method if it is available. Available methods include:
             - name:some_name
             - id:some_id
             - link:link_text
             - partial link text:partial_link_text
             - class:class_of_element
             - tag:element_tag
             - css:css_selector

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
           - ``get_multiples``: If get_web_element should return a list of elements that match the given locator. False by default.
        """

        # Timeout for element to be found may come from the timeout given to this method or the default_timeout.
        if timeout:
            self.wait._timeout = timeout
        else:
            self.wait._timeout = self.default_timeout

        location_methods = {
            "name": By.NAME,
            "id": By.ID,
            "link": By.LINK_TEXT,
            "xpath": By.XPATH,
            "partial link text": By.PARTIAL_LINK_TEXT,
            "class": By.CLASS_NAME,
            "tag": By.TAG_NAME,
            "css": By.CSS_SELECTOR,
        }

        multi_element_return = {
            False: EC.visibility_of_element_located,
            True: EC.visibility_of_all_elements_located
        }

        try:
            if type(locator) is WebElement or type(locator) is MoonElement:
                results = locator
            elif locator.startswith("/"):
                results = self.wait.until(multi_element_return.get(get_multiples)((By.XPATH, locator)))
            elif locator.split(":")[0] in location_methods:
                results = self.wait.until(multi_element_return.get(get_multiples)((location_methods[locator.split(":")[0]], locator.split(":")[1])))
            else:
                results = self.wait.until(multi_element_return.get(get_multiples)((By.CSS_SELECTOR, locator)))
        except TimeoutException:
            raise TimeoutException(f"Could not find '{locator}' after {self.wait._timeout} seconds.")
        
        # Convert WebElement(s) to MoonElement(s)
        if type(results) is list:
            elements = []
            for e in results:
                elements.append(MoonElement(e))
            return elements
        else:
            return MoonElement(results)

    def click_element(self, locator, timeout=None):
        """Click the element identified by ``locator``.

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
        """
        self.get_web_element(locator=locator, timeout=timeout).click()


    def input_text(self, locator, text, timeout=None):
        """Types the given ``text`` into the text field identified by ``locator``.

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
        """
        self.get_web_element(locator=locator, timeout=timeout).send_keys(text)

    def get_web_elements(self, locator, timeout=None):
        """Returns a list of WebElement objects matching the ``locator``.

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
        """
        return self.get_web_element(locator=locator, timeout=timeout, get_multiples=True)

    def select_from_list_by_value(self, locator, value, timeout=None):
        """Select option from selection list ``locator`` by ``value``.

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``value``: The value to select from the list.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
        """
        Select(self.get_web_element(locator=locator, timeout=timeout)).select_by_value(value)
        
    def select_from_list_by_label(self, locator, label, timeout=None):
        """Select option from selection list ``locator`` by ``label``.

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``label``: The label to select from the list.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
        """
        Select(self.get_web_element(locator=locator, timeout=timeout)).select_by_visible_text(label)

    def select_from_list_by_index(self, locator, index, timeout=None):
        """Select option from selection list ``locator`` by ``index``.

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``index``: The index to select from the list.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
        """
        Select(self.get_web_element(locator=locator, timeout=timeout)).select_by_index(index) 

    def get_text(self, locator, timeout=None):
        """Returns the text value of the element identified by ``locator``.

           Arguments:
           - ``locator``: The method by which to locate a web element.
           - ``timeout``: How long to search for the element before throwing a TimeoutException.
        """
        return self.get_web_element(locator=locator, timeout=timeout).text
    
    def drag_element_to_element(self, dragged_locator: str, target_locator: str, timeout: int = None):
        """Clicks and drags one element to another element.

           Arguments:
           - ``dragged_locator`` (str): Locator for the element to be clicked and dragged.
           - ``target_locator`` (str): Locator for the element that the dragged element will be dragged to.
           - ``timeout`` (int, optional): Maximum time (in seconds) to search for both the dragged and target elements before throwing a TimeoutException.
        """
        dragged_element = self.get_web_element(locator=dragged_locator, timeout=timeout)
        target_element = self.get_web_element(locator=target_locator, timeout=timeout)

        ActionChains(self.moon_driver).drag_and_drop(dragged_element, target_element).perform()

    def drag_element_to_offset(self, dragged_locator: str, x_offset: int, y_offset: int, timeout: int = None):
        """Clicks and drags one element to another element.

           Arguments:
           - ``dragged_locator`` (str): Locator for the element to be clicked and dragged.
           - ``x_offset`` (int): The number of pixels to drag the element along the x axis.
           - ``y_offset`` (int): The number of pixels to drag the element along the y axis.
           - ``timeout`` (int, optional): Maximum time (in seconds) to search for the dragged element before throwing a TimeoutException.
        """
        dragged_element = self.get_web_element(locator=dragged_locator, timeout=timeout)

        ActionChains(self.moon_driver).drag_and_drop_by_offset(dragged_element, xoffset=x_offset, yoffset=y_offset).perform()

    def scroll_to_element(self, locator: str, timeout: int = None):
        """Scrolls an element into view.

           Arguments:
           - ``locator`` (str): Locator for the element to be scrolled into view.
           - ``timeout`` (int, optional): Maximum time (in seconds) to search for the element before throwing a TimeoutException.
        """
        element = self.get_web_element(locator=locator, timeout=timeout)

        self.moon_driver.execute_script("arguments[0].scrollIntoView();", element)

class MoonElement(WebElement):
    """Replacement object for standard WebElements.
       Inherits all attributes and methods of Selenium WebElements with additional methods. 
    """

    def __init__(self, element: WebElement):
        self.element = element
        self.driver = element.parent
        super().__init__(self.driver, element.id)

    def get_parent(self):
        """Return the parent element of a WebElement.
        """
        return MoonElement(self.driver.execute_script("return arguments[0].parentNode", self.element))
    
    def get_children(self, locator: str = "*"):
        """Returns all child elements of a WebElement.
           Arguments:
           - ``locator``: The method by which to locate children of the given WebElement.
        """
        elements = []
        for e in self.driver.execute_script(f"return arguments[0].querySelectorAll('{locator}')", self.element):
            elements.append(MoonElement(e))
        return elements


import threading
import time
from datetime import datetime
import os
import ffmpeg

class VideoThread(threading.Thread):
    """Creates a new thread that handles the creation of video recordings for browser sessions.
    """
    def __init__(self, driver, shutter_speed):
        """Receives a webdriver to create an additional thread where screenshots are taken during test execution.
           These screenshots are converted to an mp4 file when the webdriver is killed or after the moonrise test execution has completed.
        """
        threading.Thread.__init__(self)
        self.stop_event = threading.Event()
        self.driver = driver
        self.video_folder = None
        self.shutter_speed = shutter_speed
        self.start()

    def run(self):
        while not self.stop_event.is_set():
            try:
                timestamp = str(datetime.now()).replace(" ", "_").replace(":", "_")
                self.driver.save_screenshot(f"{self.video_folder}/{timestamp}.png")
                time.sleep(self.shutter_speed)
            except Exception:
                self.stop()

    def stop(self):
        self.stop_event.set()

    def create_video_from_pngs(self, output_file):

        # Get a list of PNG files in the folder
        png_files = [file for file in os.listdir(self.video_folder) if file.endswith('.png')]

        # Sort the files in ascending order based on their names
        png_files.sort()

        # Set up the FFmpeg input file list
        input_file_list = os.path.join(self.video_folder, 'input.txt')
        with open(input_file_list, 'w') as f:
            for i, png_file in enumerate(png_files):
                file_path = os.path.join(self.video_folder, png_file)
                f.write(f"file '{file_path}'\nduration 0.1\n")
        
        ffmpeg.input(input_file_list, format='concat', safe=0).output(
            output_file,
            r=10,  # Set framerate to 10 frames per second (adjust as needed)
            start_number=0  # Set the start number of the input files
        ).overwrite_output().run()



from moonrise.Moon_Browser import MoonBrowser
from moonrise.Moon_Methods import MoonMethods
from moonrise.Base_Test import BaseTest

class Moonrise(MoonBrowser, MoonMethods, BaseTest):
    """Moonrise is a test suite creation toolset with additional quality of life upgrades for using the [selenium](https://www.selenium.dev/) test framework.

       Notable features:
       - Test Suite and Test Case organization
       - Browser session generation with selenium
       - Web element lookup methods with built-in dynamic waits for elements to become available
       - Test report generation

       Moonrise is designed for ultimate ease of use while still giving access to the power of the selenium framework. 
       By simply extending the Moonrise class, any Python class becomes a test suite, capable of executing automated test cases and generating test reports.

       ### Example ###
        from moonrise import Moonrise

        class ExampleSuite(Moonrise):

            @Moonrise.test
            def example_test(self):
                self.log_to_report("this is a test")

            
       The easiest way to use Moonrise is through the command line. The CLI offers a broad range of filtering options:

       - Use `moonrise (a folder containing python modules)` to execute all Moonrise tests and suites within that folder.
       - Use `moonrise (a python file)` to target a specific file that may contain Moonrise tests and suites.
       - Use `test:(test name)` to target specific tests. This keyword may be used more than once in a command and can apply to similarly named tests across multiple suites and python files.
       - Use `suite:(suite name)` to target specific test suites. This keyword may be used more than once in a command and can apply to similarly named suites across multiple python files.
    """